@@ -1,9 +1,11 @@
 """
-Base class describing a model.
+Base class describing a model
 
 .. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
 """
 
+from __future__ import annotations
+
 import argparse
 import functools
 import importlib.util
@@ -17,12 +19,11 @@ from typing import (
     TYPE_CHECKING,
     Any,
     Callable,
-    Dict,
-    Optional,
+    Literal,
     Sequence,
-    Type,
     TypeVar,
-    Union,
+    get_args,
+    get_origin,
 )
 
 from .parameters import (
@@ -32,7 +33,7 @@ from .parameters import (
     Parameter,
     Parameterized,
 )
-from .state import ObjectState, StateBase
+from .storage import MemoryStorage, ModeType, StorageGroup, open_storage  # type: ignore
 
 if TYPE_CHECKING:
     from .results import Result  # @UnusedImport
@@ -41,18 +42,17 @@ if TYPE_CHECKING:
 class ModelBase(Parameterized, metaclass=ABCMeta):
     """base class for describing models"""
 
-    name: Optional[str] = None
+    name: str | None = None
     """str: the name of the model"""
-    description: Optional[str] = None
+    description: str | None = None
     """str: a longer description of the model"""
-    state_cls: Type[StateBase] = ObjectState
-    """type: the class that will contain the model results"""
 
     def __init__(
         self,
-        parameters: Optional[Dict[str, Any]] = None,
-        output: Optional[str] = None,
+        parameters: dict[str, Any] | None = None,
+        output: str | None = None,
         *,
+        mode: ModeType = "insert",
         strict: bool = False,
     ):
         """initialize the parameters of the object
@@ -64,26 +64,52 @@ class ModelBase(Parameterized, metaclass=ABCMeta):
                 :meth:`~Parameterized.get_parameters` or displayed by calling
                 :meth:`~Parameterized.show_parameters`.
             output (str):
-                Path to write the output file
+                Path where the output file will be written. The output will be written
+                using :mod:`~modelrunner.storage` and might contain two groups: `result`
+                to which the final result of the model is written, and `data`, which
+                can contain extra information that is written using
+                :attr:`~ModelBase.storage`.
+            mode (str or :class:`~modelrunner.storage.access_modes.ModeType`):
+                The file mode with which the storage is accessed, which determines the
+                allowed operations. Common options are "read", "full", "append", and
+                "truncate".
             strict (bool):
                 Flag indicating whether parameters are strictly interpreted. If `True`,
                 only parameters listed in `parameters_default` can be set and their type
                 will be enforced.
         """
         super().__init__(parameters, strict=strict)
-        self.output = output
+        self.output = output  # TODO: also allow already opened storages
+        self.mode = mode
+        self._storage: open_storage | None = None
         self._logger = logging.getLogger(self.__class__.__name__)
 
+    @property
+    def storage(self) -> StorageGroup:
+        """:class:`StorageGroup`: Storage to which data can be written"""
+        if self._storage is None:
+            if self.output is None:
+                raise RuntimeError("Output file needs to be specified")
+            self._storage = open_storage(self.output, mode=self.mode)
+            return self._storage.create_group("data")
+        else:
+            return self._storage.open_group("data")
+
+    def close(self) -> None:
+        """close any opened storages"""
+        if self._storage is not None:
+            self._storage.close()
+        self._storage = None
+
     @abstractmethod
     def __call__(self):
-        """main method calculating the result"""
-        pass
+        """main method calculating the result. Needs to be specified by sub-class"""
 
-    def get_result(self, state: Optional[StateBase] = None) -> "Result":
+    def get_result(self, data: Any = None) -> Result:
         """get the result as a :class:`~model.Result` object
 
         Args:
-            state:
+            data:
                 The result data. If omitted, the model is run to obtain results
 
         Returns:
@@ -91,40 +117,37 @@ class ModelBase(Parameterized, metaclass=ABCMeta):
         """
         from .results import Result  # @Reimport
 
-        if state is None:
-            state_data = self()
-            state = self.state_cls(state_data)
-        elif isinstance(state, Result):
-            return state
+        if data is None:
+            data = self()
 
         info = {"time": datetime.now().strftime("%Y-%m-%d %H:%M:%S")}
-        return Result(self, state, info=info)
+        return Result(self, data, info=info)
 
-    def write_result(self, output: Optional[str] = None, result=None) -> None:
+    def write_result(self, result: Result | None = None) -> None:
         """write the result to the output file
 
         Args:
-            output (str):
-                File where the output will be written to. If omitted self.output will be
-                used. If self.output is also None, an error will be thrown.
             result:
                 The result data. If omitted, the model is run to obtain results
         """
-        if output is None:
-            output = self.output
-        if output is None:
-            raise RuntimeError("output file needs to be specified")
+        from .results import Result  # @Reimport
+
+        if self.output is None:
+            raise RuntimeError("Output file needs to be specified")
 
         if result is None:
             result = self.get_result()
-        else:
-            from .results import Result  # @Reimport
+        elif not isinstance(result, Result):
+            raise TypeError(f"result has type {result.__class__} instead of `Result`")
 
-            assert isinstance(result, Result)
-        result.to_file(output)
+        if self._storage is not None:
+            # reuse the opened storage
+            result.to_file(self._storage, loc="result")
+        else:
+            result.to_file(self.output, loc="result", mode=self.mode)
 
     @classmethod
-    def _prepare_argparser(cls, name: Optional[str] = None) -> argparse.ArgumentParser:
+    def _prepare_argparser(cls, name: str | None = None) -> argparse.ArgumentParser:
         """create argument parser for setting parameters of this model
 
         Args:
@@ -163,8 +186,8 @@ class ModelBase(Parameterized, metaclass=ABCMeta):
 
     @classmethod
     def from_command_line(
-        cls, args: Optional[Sequence[str]] = None, name: Optional[str] = None
-    ) -> "ModelBase":
+        cls, args: Sequence[str] | None = None, name: str | None = None
+    ) -> ModelBase:
         """create model from command line parameters
 
         Args:
@@ -203,8 +226,8 @@ class ModelBase(Parameterized, metaclass=ABCMeta):
 
     @classmethod
     def run_from_command_line(
-        cls, args: Optional[Sequence[str]] = None, name: Optional[str] = None
-    ) -> "Result":
+        cls, args: Sequence[str] | None = None, name: str | None = None
+    ) -> Result:
         """run model using command line parameters
 
         Args:
@@ -223,15 +246,19 @@ class ModelBase(Parameterized, metaclass=ABCMeta):
         result = mdl.get_result()
         if mdl.output:
             # write the results to a file
-            mdl.write_result(output=None, result=result)
+            mdl.write_result(result=result)
         else:
             # display the results on stdout
-            print(result.state._to_simple_objects()["data"])
+            storage = MemoryStorage()
+            result.to_file(storage)
+
+        # close the output file
+        mdl.close()
 
         return result
 
     @property
-    def _state_attributes(self) -> Dict[str, Any]:
+    def _state_attributes(self) -> dict[str, Any]:
         """dict: information about the element state, which does not change in time"""
         return {
             "class": self.__class__.__name__,
@@ -241,7 +268,7 @@ class ModelBase(Parameterized, metaclass=ABCMeta):
         }
 
 
-_DEFAULT_MODEL: Union[Callable, ModelBase, None] = None
+_DEFAULT_MODEL: Callable | ModelBase | None = None
 """stores the default model that will be used automatically"""
 
 
@@ -286,7 +313,7 @@ def cleared_default_model(func: TFunc) -> TFunc:
     return inner  # type: ignore
 
 
-def make_model_class(func: Callable, *, default: bool = False) -> Type[ModelBase]:
+def make_model_class(func: Callable, *, default: bool = False) -> type[ModelBase]:
     """create a model from a function by interpreting its signature
 
     Args:
@@ -299,18 +326,32 @@ def make_model_class(func: Callable, *, default: bool = False) -> Type[ModelBase
         :class:`ModelBase`: A subclass of ModelBase, which encompasses `func`
     """
     # determine the parameters of the function
+    provide_storage = False
     parameters_default = []
     for name, param in inspect.signature(func).parameters.items():
-        if param.annotation is param.empty:
-            cls = object
+        if name == "storage":
+            # treat this parameter specially and provide access to a storage object
+            provide_storage = True
         else:
-            cls = param.annotation
-        if param.default is param.empty:
-            default_value = NoValue
-        else:
-            default_value = param.default
+            # all remaining parameters are treated as model parameters
+            if param.annotation is param.empty:
+                cls = object
+                choices = None
+            elif get_origin(param.annotation) is Literal:
+                cls = object
+                choices = get_args(param.annotation)
+            else:
+                cls = param.annotation
+                choices = None
+            if param.default is param.empty:
+                default_value = NoValue
+            else:
+                default_value = param.default
 
-        parameters_default.append(Parameter(name, default_value=default_value, cls=cls))
+            parameter = Parameter(
+                name, default_value=default_value, cls=cls, choices=choices
+            )
+            parameters_default.append(parameter)
 
     def __call__(self, *args, **kwargs):
         """call the function preserving the original signature"""
@@ -329,6 +370,9 @@ def make_model_class(func: Callable, *, default: bool = False) -> Type[ModelBase
                 raise TypeError(f"Model missing required argument: '{name}'")
             parameters[name] = param_value
 
+        if provide_storage:
+            parameters["storage"] = self.storage
+
         return func(**parameters)
 
     args = {
@@ -346,8 +390,10 @@ def make_model_class(func: Callable, *, default: bool = False) -> Type[ModelBase
 
 def make_model(
     func: Callable,
-    parameters: Optional[Dict[str, Any]] = None,
+    parameters: dict[str, Any] | None = None,
+    output: str | None = None,
     *,
+    mode: ModeType = "insert",
     default: bool = False,
 ) -> ModelBase:
     """create model from a function and a dictionary of parameters
@@ -357,6 +403,12 @@ def make_model(
             The function that will be turned into a Model
         parameters (dict):
             Paramter values with which the model is initialized
+        output (str):
+            Path where the output file will be written.
+        mode (str or :class:`~modelrunner.storage.access_modes.ModeType`):
+            The file mode with which the storage is accessed, which determines the
+            allowed operations. Common options are "read", "full", "append", and
+            "truncate".
         default (bool):
             If True, set this model as the default one for the current script
 
@@ -364,12 +416,12 @@ def make_model(
         :class:`ModelBase`: An instance of a subclass of ModelBase encompassing `func`
     """
     model_class = make_model_class(func, default=default)
-    return model_class(parameters)
+    return model_class(parameters, output=output, mode=mode)
 
 
 def run_function_with_cmd_args(
-    func: Callable, args: Optional[Sequence[str]] = None, *, name: Optional[str] = None
-) -> "Result":
+    func: Callable, args: Sequence[str] | None = None, *, name: str | None = None
+) -> Result:
     """create model from a function and obtain parameters from command line
 
     Args:
@@ -388,7 +440,7 @@ def run_function_with_cmd_args(
 
 
 @cleared_default_model
-def run_script(script_path: str, model_args: Sequence[str]) -> "Result":
+def run_script(script_path: str, model_args: Sequence[str]) -> Result:
     """helper function that runs a model script
 
     The function detects models automatically by trying several methods until one yields
@@ -416,7 +468,7 @@ def run_script(script_path: str, model_args: Sequence[str]) -> "Result":
     filename = os.path.basename(script_path)
     spec = importlib.util.spec_from_file_location("model_code", script_path)
     if spec is None:
-        raise IOError(f"Could not find job script `{script_path}`")
+        raise OSError(f"Could not find job script `{script_path}`")
     model_code = importlib.util.module_from_spec(spec)
     spec.loader.exec_module(model_code)  # type: ignore
 
