@@ -10,6 +10,7 @@ One aim is to allow easy management of inheritance of parameters.
    DeprecatedParameter
    HideParameter
    Parameterized
+   get_all_parameters
 
 .. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
 """
@@ -17,14 +18,15 @@ One aim is to allow easy management of inheritance of parameters.
 from __future__ import annotations
 
 import copy
-import importlib
 import logging
 import warnings
 from dataclasses import dataclass, field
-from typing import Any, Callable, Dict, Iterator, List, Optional, Type, Union
+from typing import Any, Callable, Container, Dict, Iterator, List, Optional, Union
 
 import numpy as np
 
+from .utils import hybridmethod, import_class
+
 
 class NoValueType:
     """special value to indicate no value for a parameter"""
@@ -54,25 +56,6 @@ def auto_type(value):
         return float_val
 
 
-def import_class(identifier: str):
-    """import a class or module given an identifier
-
-    Args:
-        identifier (str):
-            The identifier can be a module or a class. For instance, calling the
-            function with the string `identifier == 'numpy.linalg.norm'` is
-            roughly equivalent to running `from numpy.linalg import norm` and
-            would return a reference to `norm`.
-    """
-    module_path, _, class_name = identifier.rpartition(".")
-    if module_path:
-        module = importlib.import_module(module_path)
-        return getattr(module, class_name)
-    else:
-        # this happens when identifier does not contain a dot
-        return importlib.import_module(class_name)
-
-
 @dataclass
 class Parameter:
     """class representing a single parameter
@@ -81,12 +64,20 @@ class Parameter:
         name (str):
             The name of the parameter
         default_value:
-            The default value
+            The default value of the parameter
         cls:
-            The type of the parameter, which is used for conversion
+            The type of the parameter, which is used for conversion. The conversion and
+            parsing of values can be disabled by using the default class `object`.
         description (str):
-            A string describing the impact of this parameter. This
-            description appears in the parameter help
+            A string describing the impact of this parameter. This description appears
+            in the parameter help.
+        choices (container):
+            A list or set of values that the parameter can take. Values (including the
+            default value) that are not in this list will be rejected. Note that values
+            are check after they have been converted by `cls`, so specifying `cls` is
+            particularly important to convert command line parameters from strings.
+        required (bool):
+            Whether the parameter is required
         hidden (bool):
             Whether the parameter is hidden in the description summary
         extra (dict):
@@ -95,10 +86,17 @@ class Parameter:
 
     name: str
     default_value: Any = None
-    cls: Union[Type, Callable] = object
+    cls: type | Callable = object
     description: str = ""
+    choices: Container | None = None
+    required: bool = False
     hidden: bool = False
-    extra: Dict[str, Any] = field(default_factory=dict)
+    extra: dict[str, Any] = field(default_factory=dict)
+
+    def _check_value(self, value) -> None:
+        """checks whether the value is acceptable"""
+        if value is not None and self.choices is not None and value not in self.choices:
+            raise ValueError(f"Default value `{value}` not in `{self.choices}`")
 
     def __post_init__(self):
         """check default values and cls"""
@@ -110,9 +108,12 @@ class Parameter:
                 converted_value = self.cls(self.default_value)
             except TypeError as err:
                 raise TypeError(
-                    f"Parameter {self.name} has invalid default: {self.default_value}"
+                    f"Parameter {self.name} of type {self.cls} has invalid default "
+                    f"value: {self.default_value}"
                 ) from err
 
+            self._check_value(converted_value)
+
             if isinstance(converted_value, np.ndarray):
                 # numpy arrays are checked for each individual value
                 valid_default = np.allclose(
@@ -140,6 +141,8 @@ class Parameter:
             "default_value": self.convert(),
             "cls": self.cls.__module__ + "." + self.cls.__name__,
             "description": self.description,
+            "choices": self.choices,
+            "required": self.required,
             "hidden": self.hidden,
             "extra": self.extra,
         }
@@ -156,8 +159,8 @@ class Parameter:
         return self.description.split(". ", 1)[0]
 
     def convert(self, value=NoValue, *, strict: bool = True):
-        """converts a `value` into the correct type for this parameter. If
-        `value` is not given, the default value is converted.
+        """converts a `value` into the correct type for this parameter. If `value` is
+        not given, the default value is converted.
 
         Note that this does not make a copy of the values, which could lead to
         unexpected effects where the default value is changed by an instance.
@@ -191,6 +194,7 @@ class Parameter:
                     ) from err
                 # else: just return the value unchanged
 
+        self._check_value(value)
         return value
 
     def _argparser_add(self, parser):
@@ -202,7 +206,12 @@ class Parameter:
                 description = f"Parameter `{self.name}`"
 
             arg_name = "--" + self.name
-            kwargs = {"default": self.default_value, "help": description}
+            kwargs = {
+                "required": self.required,
+                "choices": self.choices,
+                "default": self.default_value,
+                "help": description,
+            }
 
             if self.cls is bool:
                 # parameter is a boolean that we want to adjust
@@ -268,35 +277,8 @@ class HideParameter:
         pass
 
 
-class hybridmethod:
-    """
-    descriptor that can be used as a decorator to allow calling a method both
-    as a classmethod and an instance method
-
-    Adapted from https://stackoverflow.com/a/28238047
-    """
-
-    def __init__(self, fclass, finstance=None, doc=None):
-        self.fclass = fclass
-        self.finstance = finstance
-        self.__doc__ = doc or fclass.__doc__
-        # support use on abstract base classes
-        self.__isabstractmethod__ = bool(getattr(fclass, "__isabstractmethod__", False))
-
-    def classmethod(self, fclass):
-        return type(self)(fclass, self.finstance, None)
-
-    def instancemethod(self, finstance):
-        return type(self)(self.fclass, finstance, self.__doc__)
-
-    def __get__(self, instance, cls):
-        if instance is None or self.finstance is None:
-            # either bound to the class, or no instance method available
-            return self.fclass.__get__(cls, None)
-        return self.finstance.__get__(instance, cls)
-
-
 ParameterListType = List[Union[Parameter, HideParameter]]
+ParameterInputType = Optional[Dict[str, Any]]
 
 
 class Parameterized:
@@ -306,12 +288,10 @@ class Parameterized:
     """list: parameters (with default values) of this subclass"""
     _parameters_default_full: ParameterListType = []
     """list: all parameters (including those of parent classes)"""
-    _subclasses: Dict[str, Type[Parameterized]] = {}
+    _subclasses: dict[str, type[Parameterized]] = {}
     """dict: a dictionary of all classes inheriting from `Parameterized`"""
 
-    def __init__(
-        self, parameters: Optional[Dict[str, Any]] = None, *, strict: bool = True
-    ):
+    def __init__(self, parameters: ParameterInputType = None, *, strict: bool = True):
         """initialize the parameters of the object
 
         Args:
@@ -347,7 +327,7 @@ class Parameterized:
             ]
 
         # combine parameters with those of the parent class
-        parameters_default: Dict[str, Parameter] = {}
+        parameters_default: dict[str, Parameter] = {}
         for p in cls._parameters_default_full + cls.parameters_default:
             if isinstance(p, HideParameter):
                 if p.name in parameters_default:
@@ -387,20 +367,22 @@ class Parameterized:
         include_hidden: bool = False,
         include_deprecated: bool = False,
         sort: bool = True,
-    ) -> Dict[str, Parameter]:
+    ) -> dict[str, Parameter]:
         """return a dictionary of parameters that the class supports
 
         Args:
-            include_hidden (bool): Include hidden parameters
-            include_deprecated (bool): Include deprecated parameters
-            sort (bool): Return ordered dictionary with sorted keys
+            include_hidden (bool):
+                Include hidden parameters
+            include_deprecated (bool):
+                Include deprecated parameters
+            sort (bool):
+                Return ordered dictionary with sorted keys
 
         Returns:
-            dict: a dictionary of instance of :class:`Parameter` with their
-            names as keys.
+            dict: a dictionary mapping names to instances of :class:`Parameter`
         """
         # collect the parameters from the class hierarchy
-        parameters: Dict[str, Parameter] = {}
+        parameters: dict[str, Parameter] = {}
         for p in cls._parameters_default_full:
             if isinstance(p, HideParameter):
                 if include_hidden:
@@ -432,24 +414,28 @@ class Parameterized:
     @classmethod
     def _parse_parameters(
         cls,
-        parameters: Optional[Dict[str, Any]] = None,
+        parameters: ParameterInputType = None,
+        *,
         check_validity: bool = True,
         allow_hidden: bool = True,
         include_deprecated: bool = False,
-    ) -> Dict[str, Any]:
-        """parse parameters
+    ) -> dict[str, Any]:
+        """parse parameters from a given dictionary
 
         Args:
             parameters (dict):
                 A dictionary of parameters that will be parsed.
             check_validity (bool):
                 Determines whether a `ValueError` is raised if there are keys in
-                parameters that are not in the defaults. If `False`, additional
-                items are simply stored in `self.parameters`
+                parameters that are not in the defaults. If `False`, additional items
+                are simply stored in `self.parameters`
             allow_hidden (bool):
                 Allow setting hidden parameters
             include_deprecated (bool):
                 Include deprecated parameters
+
+        Returns:
+            dict: The parsed parameters
         """
         if parameters is None:
             parameters = {}
@@ -462,10 +448,12 @@ class Parameterized:
         )
 
         # initialize parameters with default ones from all parent classes
-        result: Dict[str, Any] = {}
+        result: dict[str, Any] = {}
         for name, param_obj in param_objs.items():
             if not allow_hidden and param_obj.hidden:
                 continue  # skip hidden parameters
+            if param_obj.required and name not in parameters:
+                raise ValueError(f"Require parameter `{name}`")
             # take value from parameters or set default value
             value = parameters.pop(name, NoValue)
             # convert parameter to correct type
@@ -504,9 +492,9 @@ class Parameterized:
         show_hidden: bool = False,
         show_deprecated: bool = False,
         short_description: bool = False,
-        parameter_values: Optional[Dict[str, Any]] = None,
-        template: Optional[str] = None,
-        template_object: Optional[str] = None,
+        parameter_values: ParameterInputType = None,
+        template: str | None = None,
+        template_object: str | None = None,
     ) -> Iterator[str]:
         """private method showing all parameters in human readable format
 
@@ -522,8 +510,8 @@ class Parameterized:
             short_description (bool):
                 Whether to show a shortended version of the description
             parameter_values (dict):
-                A dictionary with values to show. Parameters not in this
-                dictionary are shown with their default value.
+                A dictionary with values to show. Parameters not in this dictionary are
+                shown with their default value.
 
         All flags default to `False`.
         """
@@ -546,9 +534,9 @@ class Parameterized:
             data = {
                 "name": param.name,
                 "type": param.cls.__name__,
-                "description": param.short_description
-                if short_description
-                else param.description,
+                "description": (
+                    param.short_description if short_description else param.description
+                ),
             }
 
             # determine the value to show
@@ -566,7 +554,7 @@ class Parameterized:
     @hybridmethod
     def show_parameters(  # @NoSelf
         cls,
-        description: bool = False,  # @NoSelf
+        description: bool = False,
         sort: bool = False,
         show_hidden: bool = False,
         show_deprecated: bool = False,
@@ -596,7 +584,7 @@ class Parameterized:
     @show_parameters.instancemethod  # type: ignore
     def show_parameters(
         self,
-        description: bool = False,  # @NoSelf
+        description: bool = False,
         sort: bool = False,
         show_hidden: bool = False,
         show_deprecated: bool = False,
@@ -614,8 +602,8 @@ class Parameterized:
             show_deprecated (bool):
                 Flag determining whether deprecated parameters are shown
             default_value (bool):
-                Flag determining whether the default values or the current
-                values are shown
+                Flag determining whether the default values or the current values are
+                shown
 
         All flags default to `False`.
         """
@@ -629,14 +617,13 @@ class Parameterized:
             print(line)
 
 
-def get_all_parameters(data: str = "name") -> Dict[str, Any]:
+def get_all_parameters(data: str = "name") -> dict[str, Any]:
     """get a dictionary with all parameters of all registered classes
 
     Args:
         data (str):
-            Determines what data is returned. Possible values are 'name',
-            'value', or 'description', to return the respective information
-            about the parameters.
+            Determines what data is returned. Possible values are 'name', 'value', or
+            'description', to return the respective information about the parameters.
     """
     result = {}
     for cls_name, cls in Parameterized._subclasses.items():
