@@ -1,25 +1,32 @@
 """
+Provides functions for submitting models as jobs
+
 .. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
 """
 
+from __future__ import annotations
+
 import errno
 import itertools
 import json
 import logging
 import os
-import pipes
+import shlex
 import subprocess as sp
 import sys
 import warnings
 from pathlib import Path
-from typing import Any, Dict, Iterable, Optional, Tuple, Union
+from typing import Any, Iterable, Literal
 
 from tqdm.auto import tqdm
 
+from ..config import Config
+from ..parameters import Parameter
+
 
 def escape_string(obj) -> str:
     """escape a string for the command line"""
-    return pipes.quote(str(obj))
+    return shlex.quote(str(obj))
 
 
 def ensure_directory_exists(folder):
@@ -33,15 +40,67 @@ def ensure_directory_exists(folder):
             raise
 
 
-def get_job_name(
-    base: str, args: Optional[Dict[str, Any]] = None, length: int = 7
-) -> str:
+DEFAULT_CONFIG = [
+    Parameter(
+        "python_bin",
+        "sys.executable",
+        str,
+        "Path to the python executable to be used. The special value `sys.executable` "
+        "uses the value returned by `sys.executable`.",
+    ),
+    Parameter("num_threads", 1, int, "The number of threads to be used"),
+    Parameter(
+        "method",
+        "qsub",
+        str,
+        "The job submission method.",
+        choices=["background", "foreground", "qsub", "srun"],
+    ),
+    Parameter(
+        "partition", "", str, "The partition to which the job will be submitted."
+    ),
+]
+
+
+def get_config(
+    config: str | dict[str, Any] | None = None, *, load_user_config: bool = True
+) -> Config:
+    """create the job configuration
+
+    Args:
+        config (str or dict):
+            Configuration settings that will be used to update the default config
+        load_user_config (bool):
+            Determines whether the file `~/.modelrunner` is loaded as a YAML document to
+            provide user-defined settings.
+
+    Returns:
+        :class:`~modelrunner.config.Config`: the established configuration
+    """
+    c = Config(DEFAULT_CONFIG)
+
+    if load_user_config:
+        path = Path.home() / ".modelrunner"
+        if path.is_file():
+            c.load(path)
+
+    if isinstance(config, str):
+        c.update(json.loads(config))
+    elif config is not None:
+        c.update(config)
+    return c
+
+
+def get_job_name(base: str, args: dict[str, Any] | None = None, length: int = 7) -> str:
     """create a suitable job name
 
     Args:
-        base (str): The stem of the job name
-        args (dict): Parameters to include in the job name
-        length (int): Length of the abbreviated parameter name
+        base (str):
+            The stem of the job name
+        args (dict):
+            Parameters to include in the job name
+        length (int):
+            Length of the abbreviated parameter name
 
     Returns:
         str: A suitable job name
@@ -59,19 +118,25 @@ def get_job_name(
     return res
 
 
+OverwriteStrategyType = Literal[
+    "error", "warn_skip", "silent_skip", "overwrite", "silent_overwrite"
+]
+
+
 def submit_job(
-    script: Union[str, Path],
-    output: Union[str, Path, None] = None,
+    script: str | Path,
+    output: str | Path | None = None,
     name: str = "job",
-    parameters: Union[str, Dict[str, Any], None] = None,
+    parameters: str | dict[str, Any] | None = None,
+    config: str | dict[str, Any] | None = None,
     *,
-    log_folder: Union[str, Path] = "logs",
-    method: str = "qsub",
+    log_folder: str | Path | None = None,
+    method: str = "auto",
     use_modelrunner: bool = True,
-    template: Union[str, Path, None] = None,
-    overwrite_strategy: str = "error",
+    template: str | Path | None = None,
+    overwrite_strategy: OverwriteStrategyType = "error",
     **kwargs,
-) -> Tuple[str, str]:
+) -> tuple[str, str]:
     """submit a script to the cluster queue
 
     Args:
@@ -84,11 +149,18 @@ def submit_job(
         parameters (str or dict):
             Parameters for the script, either as a python dictionary or a string
             containing a JSON-encoded dictionary.
+        config (str or dict):
+            Configuration for the job, which determines how the job is run. Can be either
+            a python dictionary or a string containing a JSON-encoded dictionary.
         log_folder (str of :class:`~pathlib.Path`):
-            Path to the logging folder
+            Path to the logging folder. If omitted, the default of the template is used,
+            which typically sends data to stdout for local scripts (which is thus
+            captured and returned by this function) or writes log files to the current
+            working directory for remote jobs.
         method (str):
-            Specifies the submission method. Currently `background`, `foreground`, 'srun' and
-            `qsub` are supported.
+            Specifies the submission method. Currently `background`, `foreground`,
+            'srun', and `qsub` are supported. The special value `auto` reads the method
+            from the `config` argument.
         use_modelrunner (bool):
             If True, `script` is envoked with the modelrunner library, e.g. by calling
             `python -m modelrunner {script}`.
@@ -98,40 +170,50 @@ def submit_job(
         overwrite_strategy (str):
             Determines what to do when files already exist. Possible options include
             `error`, `warn_skip`, `silent_skip`, `overwrite`, and `silent_overwrite`.
-        **kwargs:
-            Extra arguments are forwarded as template variables to the script. A typical
-            use case is specifying the partition in a `qsub` script, which can be done
-            by giving the `partition` parameter here
 
     Returns:
-        tuple: The result `(stdout, stderr)` of the submission call
+        tuple: The result `(stdout, stderr)` of the submission call. These two strings
+            can contain the output from the actual scripts that is run when `log_folder`
+            is `None`.
     """
     from jinja2 import Template
 
     logger = logging.getLogger("modelrunner.submit_job")
 
+    # prepare job configuration
+    configuration = get_config(config)
+    if kwargs:
+        # deprecated since 2024-01-03
+        warnings.warn("kwargs are deprecated. Use `config` instead", DeprecationWarning)
+        for k, v in kwargs.items():
+            configuration[k] = v
+    if configuration["python_bin"] == "sys.executable":
+        configuration["python_bin"] = sys.executable
+
+    # determine the submission method
+    if method == "auto":
+        method = configuration["method"]
+
+    # load the correct template
     if template is None:
-        template_path = Path(__file__).parent / "templates" / (method + ".template")
+        template_path = Path(__file__).parent / "templates" / (method + ".jinja")
     else:
         template_path = Path(template)
     logger.info("Load template `%s`", template_path)
-    with open(template_path, "r") as fp:
+    with open(template_path) as fp:
         script_template = fp.read()
 
     # prepare submission script
-    ensure_directory_exists(log_folder)
-
-    script_args = {
+    script_args: dict[str, Any] = {
         "PACKAGE_PATH": Path(__file__).parents[2],
-        "LOG_FOLDER": log_folder,
         "JOB_NAME": name,
         "MODEL_FILE": escape_string(script),
         "USE_MODELRUNNER": use_modelrunner,
-        "PYTHON_BIN": sys.executable,
-        "DEFAULT_QUEUE": "teutates.q",
+        "CONFIG": configuration,
     }
-    for k, v in kwargs.items():
-        script_args[k.upper()] = v
+    if log_folder is not None:
+        ensure_directory_exists(log_folder)
+        script_args["LOG_FOLDER"] = log_folder
 
     # add the parameters to the job arguments
     job_args = []
@@ -164,13 +246,13 @@ def submit_job(
                 raise NotImplementedError(f"Unknown strategy `{overwrite_strategy}`")
 
             # delete old output
-            output.unlink(missing_ok=True)
+            output.unlink()
 
         # check whether output points to a directory or whether this should be a file
         if output.is_dir():
-            script_args["OUTPUT_FOLDER"] = pipes.quote(str(output))
+            script_args["OUTPUT_FOLDER"] = shlex.quote(str(output))
         else:
-            script_args["OUTPUT_FOLDER"] = pipes.quote(str(output.parent))
+            script_args["OUTPUT_FOLDER"] = shlex.quote(str(output.parent))
             job_args.append(f"--output {escape_string(output)}")
 
     else:
@@ -192,8 +274,19 @@ def submit_job(
             universal_newlines=True,
         )
 
-    elif method in {"background", "foreground"}:
-        # run job locally
+    elif method == "foreground":
+        # run job locally in the foreground, blocking further calls
+        proc = sp.Popen(
+            ["bash"],
+            stdin=sp.PIPE,
+            stdout=sp.PIPE,
+            stderr=sp.PIPE,
+            universal_newlines=True,
+            bufsize=0,  # write output immediately
+        )
+
+    elif method == "background":
+        # run job locally in the background
         proc = sp.Popen(
             ["bash"],
             stdin=sp.PIPE,
@@ -209,13 +302,14 @@ def submit_job(
 
 
 def submit_jobs(
-    script: Union[str, Path],
-    output_folder: Union[str, Path],
+    script: str | Path,
+    output_folder: str | Path,
     name_base: str = "job",
-    parameters: Union[str, Dict[str, Any], None] = None,
+    parameters: str | dict[str, Any] | None = None,
+    config: str | dict[str, Any] | None = None,
     *,
     output_format: str = "hdf",
-    list_params: Optional[Iterable[str]] = None,
+    list_params: Iterable[str] | None = None,
     **kwargs,
 ) -> int:
     """submit many jobs of the same script with different parameters to the cluster
@@ -232,6 +326,9 @@ def submit_jobs(
             containing a JSON-encoded dictionary. All combinations of parameter values
             that are iterable and not strings and not part of `keep_list` are submitted
             as separate jobs.
+        config (str or dict):
+            Configuration for the job, which determines how the job is run. Can be either
+            a python dictionary or a string containing a JSON-encoded dictionary.
         output_format (str):
             File extension determining the output format
         list_params (list):
@@ -278,6 +375,8 @@ def submit_jobs(
         params.update(p_job)
         name = get_job_name(name_base, p_job)
         output = Path(output_folder) / f"{name}{output_format}"
-        submit_job(script, output=output, name=name, parameters=params, **kwargs)
+        submit_job(
+            script, output=output, name=name, parameters=params, config=config, **kwargs
+        )
 
     return len(p_vary_list)
